import logging
from collections import defaultdict
from typing import Dict, Generator, List, Optional

from django.contrib.contenttypes.models import ContentType
from rest_framework import filters
from rest_framework.pagination import CursorPagination, LimitOffsetPagination
from rest_framework.request import Request
from rest_framework.reverse import reverse

from bridger.enums import Button, WidgetType
from bridger.filters import DjangoFilterBackend
from bridger.serializers import ListSerializer, RepresentationSerializer

from .metadata import BridgerMetaData
from .utils import ilen

logger = logging.getLogger(__name__)


class MetadataMixin:
    """
    A mixin that adds the BridgerMetaData to a View with all functionalities
    needed to properly create a valid OPTION request:

    === IDENTIFIER ===
    The identifier is a unique string which identifies this endpoint. This can be used
    for a frontend caching mechanism.

    === ENDPOINT ===
    The endpoints display the url where something is found. Three different endpoints
    are being supported:
    - instance_endpoint
    - list_endpoint
    - new_instance_endpoint

    === BUTTONS ===
    The buttons describe what kind of buttons are appropriate for this endpoint.

    - buttons: Default buttons which are displayed
       - Options: [save, save_and_new, save_and_close, delete, refresh, new]
    - custom_buttons: Custom Buttons that do stuff
    - custom_instance_buttons: Custom buttons which are tied to additional resources of an instance


    === MESSAGES ===
    Messages that should be conveyed through the OPTIONS request

    === DISPLAY ===
    The displays describe how a list or an instance should be displayed. Essentially,
    they describe a table layout for lists and a form for instances.

    === TITLES ===
    Returns the Widget Titles for Instance, List and Create Widgets

    """

    metadata_class = BridgerMetaData

    def get_widget_type(self, request: Request) -> str:
        widget_type = getattr(self, "WIDGET_TYPE", None)

        return widget_type or (
            WidgetType.LIST.value
            if "pk" not in self.kwargs
            else WidgetType.INSTANCE.value
        )

    def get_identifier(self, request: Request) -> str:
        identifier = getattr(self, "IDENTIFIER", None)
        if not identifier and hasattr(self, "get_serializer_class"):
            ct = ContentType.objects.get_for_model(
                self.get_serializer_class().Meta.model
            )
            return f"{ct.app_label}:{ct.model}"
        assert identifier, "Each Endpoint needs an Identifier"
        return identifier

    def get_endpoint(self, request: Request) -> Optional[str]:
        if hasattr(self, "ENDPOINT") and self.ENDPOINT is not None:
            return reverse(self.ENDPOINT, request=request)
        return None

    def get_list_endpoint(self, request: Request) -> Optional[str]:
        if hasattr(self, "LIST_ENDPOINT") and self.LIST_ENDPOINT is not None:
            return reverse(self.LIST_ENDPOINT, request=request)
        return self.get_endpoint(request=request)

    def get_instance_endpoint(self, request: Request) -> Optional[str]:
        if hasattr(self, "INSTANCE_ENDPOINT") and self.INSTANCE_ENDPOINT is not None:
            return reverse(self.INSTANCE_ENDPOINT, request=request)
        return self.get_endpoint(request=request)

    def get_create_endpoint(self, request: Request) -> Optional[str]:
        if hasattr(self, "CREATE_ENDPOINT") and self.CREATE_ENDPOINT is not None:
            return reverse(self.CREATE_ENDPOINT, request=request)
        return self.get_endpoint(request=request)

    def get_delete_endpoint(self, request: Request) -> Optional[str]:
        if hasattr(self, "DELETE_ENDPOINT") and self.DELETE_ENDPOINT is not None:
            return reverse(self.DELETE_ENDPOINT, request=request)
        return self.get_endpoint(request=request)

    def get_endpoints(self, request: Request, buttons: List[str]) -> Dict[str, str]:
        endpoints = dict()

        list_endpoint = self.get_list_endpoint(request=request)
        instance_endpoint = self.get_instance_endpoint(request=request)
        create_endpoint = self.get_create_endpoint(request=request)
        delete_endpoint = self.get_delete_endpoint(request=request)

        if list_endpoint:
            endpoints["list"] = list_endpoint

        if instance_endpoint:
            endpoints["instance"] = instance_endpoint

        if create_endpoint:
            endpoints["create"] = create_endpoint
        elif Button.NEW.value in buttons:
            logger.warning(
                "New Button Specified, but no create endpoint specified. New Button is removed."
            )
            buttons.remove(Button.NEW.value)

        if delete_endpoint:
            endpoints["delete"] = delete_endpoint
        elif Button.DELETE.value in buttons:
            logger.warning(
                "Delete Button Specified, but no delete endpoint specified. Delete Button is removed."
            )
            buttons.remove(Button.DELETE.value)

        return endpoints

    def get_instance_buttons(self, request: Request) -> List[str]:
        buttons = list()
        ct = None
        if hasattr(self, "get_serializer_class"):
            ct = ContentType.objects.get_for_model(
                self.get_serializer_class().Meta.model
            )

        if hasattr(self, "INSTANCE_BUTTONS"):
            return self.INSTANCE_BUTTONS
        elif ct:
            if f"{ct.app_label}.change_{ct.model}":
                buttons.append(Button.SAVE.value)
            if f"{ct.app_label}.delete_{ct.model}":
                buttons.append(Button.DELETE.value)

        return buttons

    def get_list_buttons(self, request: Request) -> List[str]:
        buttons = list()
        ct = None
        if hasattr(self, "get_serializer_class"):
            ct = ContentType.objects.get_for_model(
                self.get_serializer_class().Meta.model
            )
        if hasattr(self, "LIST_BUTTONS"):
            return self.LIST_BUTTONS
        elif ct:
            if f"{ct.app_label}.add_{ct.model}":
                buttons.append(Button.NEW.value)
        return buttons

    def get_buttons(self, request: Request) -> List[str]:
        buttons = list()

        pk = self.kwargs.get("pk", None)

        if pk:
            buttons.extend(self.get_instance_buttons(request))
        else:
            buttons.extend(self.get_list_buttons(request))

        buttons.append(Button.REFRESH.value)

        return set(buttons)

    def get_create_buttons(self, request: Request) -> List[str]:
        return getattr(
            self,
            "CREATE_BUTTONS",
            [
                Button.SAVE.value,
                Button.SAVE_AND_NEW.value,
                Button.SAVE_AND_CLOSE.value,
                Button.RESET.value,
            ],
        )

    def get_custom_buttons(self, request: Request) -> Dict[str, str]:
        return [bt.to_dict() for bt in getattr(self, "CUSTOM_BUTTONS", [])]

    def get_instance_display(self, request: Request) -> List:
        if hasattr(self, "INSTANCE_DISPLAY"):
            return self.INSTANCE_DISPLAY.to_dict()
        return []

    def get_list_display(self, request: Request) -> Dict:
        if hasattr(self, "LIST_DISPLAY"):
            return self.LIST_DISPLAY.to_dict()
        return {}

    def get_fields(self, request: Request) -> Dict:
        fields = dict()
        rs = RepresentationSerializer
        ls = ListSerializer
        field_items = self.get_serializer().fields.items()

        for name, field in filter(
            lambda f: not isinstance(f[1], (rs, ls)), field_items
        ):
            fields[name] = field.get_representation(request, name)

        for name, field in filter(lambda f: isinstance(f[1], (rs, ls)), field_items):
            representation = field.get_representation(request, name)
            fields[representation["related_key"]].update(representation)
            del fields[representation["related_key"]]["related_key"]

        return fields

    # BUTTONS
    # def get_instance_buttons(self, request):
    #     instance_buttons = getattr(self, "INSTANCE_BUTTONS", None)
    #     if instance_buttons:
    #         return instance_buttons

    #     if self.kwargs.get("pk", None):
    #         instance_buttons = ["refresh"]
    #     else:
    #         instance_buttons = ["reset"]

    #     ct = ContentType.objects.get_for_model(self.get_serializer_class().Meta.model)

    #     save_permission = f"{ct.app_label}.change_{ct.model}"
    #     delete_permission = f"{ct.app_label}.delete_{ct.model}"

    #     if request.user.has_perm(save_permission):
    #         instance_buttons.append("save")

    #     if request.user.has_perm(delete_permission) and self.kwargs.get("pk", None):
    #         instance_buttons.append("delete")

    #     return instance_buttons

    def get_preview_display(self, request: Request):
        return getattr(self, "PREVIEW_DISPLAY", "")

    def get_preview_buttons(self, request: Request):
        return [button.to_dict() for button in getattr(self, "PREVIEW_BUTTONS", [])]

    def get_custom_instance_buttons(self, request: Request):
        if "pk" in self.kwargs:
            return [
                button.to_dict()
                for button in getattr(self, "CUSTOM_INSTANCE_BUTTONS", [])
            ]
        else:
            return [
                button.to_dict()
                for button in getattr(self, "CUSTOM_LIST_INSTANCE_BUTTONS", [])
            ]

    # def get_list_buttons(self, request):
    #     list_buttons = getattr(self, "LIST_BUTTONS", None)
    #     if list_buttons:
    #         return list_buttons

    #     list_buttons = ["refresh"]

    #     ct = ContentType.objects.get_for_model(self.get_serializer_class().Meta.model)
    #     permission = f"{ct.app_label}.add_{ct.model}"

    #     if request.user.has_perm(permission):
    #         list_buttons.append("new")

    #     return list_buttons

    # PAGINATION
    def get_pagination(self, request: Request):

        pagination = (
            self.pagination_class.__name__
            if hasattr(self, "pagination_class") and self.pagination_class
            else None
        )

        return {
            "CursorPagination": "cursor",
            "LimitOffsetPagination": "page",
            None: None,
        }[pagination]

    # TITLES

    def get_instance_widget_title(self, request: Request) -> str:
        return getattr(self, "INSTANCE_WIDGET_TITLE", "")

    def get_list_widget_title(self, request: Request) -> str:
        return getattr(self, "LIST_WIDGET_TITLE", "")

    def get_create_widget_title(self, request: Request) -> str:
        return getattr(self, "CREATE_WIDGET_TITLE", "")

    def get_titles(self, request: Request) -> Dict:
        return {
            "instance": self.get_instance_widget_title(request),
            "list": self.get_list_widget_title(request),
            "create": self.get_create_widget_title(request),
        }

    # Search, Ordering, Filter
    def get_search_fields(self, request: Request) -> Generator[str, None, None]:
        if filters.SearchFilter in self.filter_backends:
            yield from self.search_fields

    def get_ordering_fields(self, request: Request) -> Generator[str, None, None]:
        if filters.OrderingFilter in self.filter_backends:
            yield from self.ordering_fields

    def get_filter_fields(self, request: Request) -> Generator[Dict, None, None]:
        if DjangoFilterBackend in self.filter_backends:
            filterset_class = DjangoFilterBackend().get_filterset_class(
                self, self.queryset
            )
            if hasattr(filterset_class, "base_filters"):
                filterset = filterset_class.base_filters.items()
                field_names = [
                    v.field_name for k, v in filterset_class.base_filters.items()
                ]
                stored_fields = dict()
                for index, (name, field) in enumerate(filterset):

                    upcomming_fields = ilen(
                        filter(
                            lambda f: f == field.field_name, field_names[index + 1 :]
                        )
                    )
                    representation = field.get_representation(request, name)

                    if upcomming_fields > 0 or field.field_name in stored_fields:
                        _representation = stored_fields.get(
                            field.field_name, representation
                        )
                        _representation["lookup_expr"].update(
                            representation["lookup_expr"]
                        )
                        stored_fields[field.field_name] = _representation

                    if upcomming_fields == 0:
                        yield field.field_name, stored_fields.get(
                            field.field_name, representation
                        )



# Generated by Django 2.2.10 on 2020-02-18 12:22

import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('bridger', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=512)),
                ('message', models.TextField(blank=True, null=True)),
                ('buttons', django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict, null=True)),
                ('timestamp_created', models.DateTimeField(auto_now_add=True)),
                ('timestamp_received', models.DateTimeField(blank=True, null=True)),
                ('timestamp_read', models.DateTimeField(blank=True, null=True)),
                ('timestamp_mailed', models.DateTimeField(blank=True, null=True)),
                ('send_type', models.CharField(choices=[('system', 'System'), ('mail', 'Mail'), ('system_and_mail', 'System and Mail')], default='system', max_length=32)),
                ('recipient', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='notifications', to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



# Generated by Django 2.2.10 on 2020-02-18 12:29

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('bridger', '0002_notification'),
    ]

    operations = [
        migrations.AlterField(
            model_name='notification',
            name='buttons',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=list, null=True),
        ),
    ]



from django.utils import timezone
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.reverse import reverse

from bridger import buttons as bt
from bridger import display as dp
from bridger import viewsets

from .models import Notification
from .serializers import NotificationModelSerializer


class NotificationModelViewSet(viewsets.ModelViewSet):
    ENDPOINT = "bridger:notification-list"

    INSTANCE_WIDGET_TITLE = "Notification: {{title}}"
    LIST_WIDGET_TITLE = "Notifications"

    LIST_BUTTONS = ["refresh"]
    INSTANCE_BUTTONS = ["refresh", "delete"]
    CREATE_BUTTONS = []

    def get_custom_buttons(self, request):
        return [
            bt.ActionButton(
                method="post",
                action_label="All notifications read.",
                endpoint=reverse(
                    "bridger:notification-mark-all-as-read", request=request
                ),
                description_fields="Do you want to mark notifications as read?",
                label="Mark all as read",
                icon="wb-icon-eye-open",
                confirm_config=bt.AdditionalButtonConfig(label="Read all"),
                cancel_config=bt.AdditionalButtonConfig(label="Cancel"),
                identifiers=["relatedmodeltest-list"],
            ).to_dict(),
            bt.ActionButton(
                method="post",
                action_label="Delete all read notifications.",
                endpoint=reverse(
                    "bridger:notification-delete-all-read", request=request
                ),
                description_fields="Do you want delete all read notifications?",
                label="Delete all read notifications",
                icon="wb-icon-trash",
                confirm_config=bt.AdditionalButtonConfig(
                    label="Delete all", level=bt.ButtonLevel.WARNING
                ),
                cancel_config=bt.AdditionalButtonConfig(
                    label="Cancel", level=bt.ButtonLevel.ERROR
                ),
                identifiers=[reverse("bridger:notification-list", request=request)],
            ).to_dict(),
        ]

    LIST_DISPLAY = dp.ListDisplay(
        fields=[
            dp.Field(key="title", label="Title"),
            dp.Field(key="timestamp_created", label="Created"),
            dp.Field(key="message", label="Message"),
        ],
        formatting=[
            dp.RowFormatting(
                column="timestamp_read",
                conditions=[
                    dp.RowIconCondition(icon="wb-icon-eye-open", condition=("∃", True)),
                ],
            ),
        ],
    )

    INSTANCE_DISPLAY = dp.InstanceDisplay(
        sections=[
            dp.Section(
                fields=dp.FieldSet(
                    fields=[
                        dp.FieldSet(fields=["timestamp_created", "timestamp_received"]),
                        dp.FieldSet(fields=["timestamp_received", "timestamp_mailed"]),
                        "title",
                        "message",
                    ]
                )
            )
        ]
    )

    queryset = Notification.objects.all()
    serializer_class = NotificationModelSerializer

    ordering_fields = ("timestamp_created",)

    def retrieve(self, request, *args, **kwargs):
        obj = self.get_object()
        if not obj.timestamp_read:
            obj.timestamp_read = timezone.now()
            obj.save()
        return super().retrieve(request, *args, **kwargs)

    def list(self, request, *args, **kwargs):
        self.get_queryset().filter(timestamp_received__isnull=True).update(
            timestamp_received=timezone.now()
        )
        return super().list(request, *args, **kwargs)

    def get_queryset(self):
        return super().get_queryset().filter(recipient=self.request.user)

    @action(methods=["POST"], detail=False)
    def mark_all_as_read(self, request, pk=None):
        Notification.objects.filter(timestamp_read__isnull=True).update(
            timestamp_read=timezone.now()
        )
        return Response({})

    @action(methods=["POST"], detail=False)
    def delete_all_read(self, request, pk=None):
        Notification.objects.filter(timestamp_read__isnull=False).delete()
        return Response({})



from django_filters.rest_framework import DjangoFilterBackend as StandardFilterBackend

from .filterset import FilterSet


class DjangoFilterBackend(StandardFilterBackend):
    filterset_base = FilterSet

    # def get_filterset_class(self, view, queryset=None):
    #     print(queryset.model)
    #     rv = super().get_filterset_class(view, queryset)
    #     print(rv)
    #     return rv



from django.urls import include, path
from django.views.generic import TemplateView

from .frontend_user_configuration import FrontendUserConfigurationModelViewSet
from .notifications.viewsets import NotificationModelViewSet
from .routers import BridgerRouter
from .views import Config, Menu, Profile

# fmt: off
router = BridgerRouter()
router.register(r"frontenduserconfiguration", FrontendUserConfigurationModelViewSet, basename="frontenduserconfiguration")
router.register(r"notification", NotificationModelViewSet, basename="notification")
# fmt: on

urlpatterns = [
    path("config/", Config.as_view(), name="config"),
    path("profile/", Profile.as_view(), name="profile"),
    path("menu/", Menu.as_view(), name="menu"),
    path("", include(router.urls)),
]


def bundled_frontend(url):
    return path(f"{url}", TemplateView.as_view(template_name="bridger/frontend.html"))



from typing import Dict

from rest_framework import serializers
from rest_framework.fields import empty
from rest_framework.relations import ManyRelatedField
from rest_framework.request import Request

from .mixins import BridgerSerializerFieldMixin
from .types import BridgerType, ReturnContentType


class BridgerManyRelatedField(ManyRelatedField):
    def __init__(self, *args, **kwargs):
        required = kwargs.get("required", True)
        if not required:
            kwargs["allow_null"] = True
        super().__init__(*args, **kwargs)

    def run_validation(self, data=empty):

        # If the data is send through form data, we need to convert the data into a proper list of ids
        if data not in [None, empty] and len(data) == 1 and "," in data[0]:
            data = data[0].split(",")

        # If the data is a list of an empty string we need to convert it (FORM DATA)
        if data not in [None, empty] and len(data) == 1 and data[0] == "":
            data = []

        # If the data is None and null is an allowed value, data needs to be set to an empty list
        if data is None and self.allow_null:
            data = []

        return super().run_validation(data)

    def get_representation(self, request: Request, field_name: str) -> Dict:
        representation = self.child_relation.get_representation(request, field_name)
        representation["multiple"] = True
        return representation


class PrimaryKeyRelatedField(
    BridgerSerializerFieldMixin, serializers.PrimaryKeyRelatedField
):

    MANY_RELATION_KWARGS = (
        "read_only",
        "write_only",
        "required",
        "default",
        "initial",
        "source",
        "label",
        "help_text",
        "style",
        "error_messages",
        "allow_empty",
        "html_cutoff",
        "html_cutoff_text",
        "allow_null",
    )

    def __init__(self, *args, **kwargs):
        self.field_type = kwargs.pop("field_type", BridgerType.SELECT.value)
        super().__init__(*args, **kwargs)

    def __new__(cls, *args, **kwargs):
        kwargs["style"] = {"base_template": "input.html"}
        return super().__new__(cls, *args, **kwargs)

    @classmethod
    def many_init(cls, *args, **kwargs):
        list_kwargs = {"child_relation": cls(*args, **kwargs)}
        for key in kwargs:
            if key in cls.MANY_RELATION_KWARGS:
                list_kwargs[key] = kwargs[key]
        return BridgerManyRelatedField(**list_kwargs)

    def run_validation(self, data=empty):
        if isinstance(data, str) and data == "null":
            data = None

        if data is empty:
            parent_model_id = self.parent.context["view"].kwargs.get(
                f"{self.field_name}_id"
            )
            if parent_model_id:
                data = parent_model_id

        return super().run_validation(data)


class ListSerializer(serializers.ListSerializer):
    """
    A Wrapper around the normal DRF ListSerializer which also return the child representation
    """

    def get_representation(self, request: Request, field_name: str) -> Dict:
        representation = self.child.get_representation(request, field_name)
        representation["multiple"] = True
        representation["related_key"] = self.source
        return representation



import logging
from typing import Dict, List

from django.conf import settings
from django.contrib.auth import get_user_model
from rest_framework.request import Request
from rest_framework.reverse import reverse

from .enums import AuthType
from .notifications.config import get_notification_config

logger = logging.getLogger(__name__)


def jwt_auth(request: Request) -> Dict:
    user_model = get_user_model()
    username_field_key = user_model.USERNAME_FIELD
    username_field_label = user_model._meta.get_field(username_field_key).verbose_name

    return {
        "type": "JWT",
        "config": {
            "token": reverse("token_obtain_pair", request=request),
            "refresh": reverse("token_refresh", request=request),
            "verify": reverse("token_verify", request=request),
            "username_field_key": username_field_key,
            "username_field_label": username_field_label,
        },
    }


def get_bridger_auth(request: Request, auth_type: AuthType = AuthType.JWT) -> Dict:
    auth_method = globals()[f"{auth_type.value.lower()}_auth"]
    return getattr(settings, "BRIDGER_AUTH", auth_method)(request)


def get_bridger_frontend_user_configuration_order() -> List:
    return getattr(
        settings, "BRIDGER_FRONTEND_USER_CONFIGURATION_ORDER", ["config__order"]
    )



import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.views import APIView

from .enums import AuthType
from .menus import default_registry
from .settings import get_bridger_auth, get_notification_config

logger = logging.getLogger(__name__)


class Menu(APIView):
    @property
    def permission_classes(self):
        bridger_auth = get_bridger_auth(None)

        if bridger_auth["type"] is None:
            return []

        return [IsAuthenticated]

    def get(self, request: Request) -> Response:
        return Response(default_registry.to_dict(request))


class Profile(APIView):
    permission_classes = [IsAuthenticated]

    def get(self, request: Request) -> Response:
        # TODO: Overwritable from withing settings.py
        return Response(
            {
                "config": reverse(
                    "bridger:frontenduserconfiguration-list", request=request
                ),
                "name": f"{request.user.first_name} {request.user.last_name}",
                "email": request.user.email,
                "image": "TODO: Here comes the URL",
            }
        )


class Config(APIView):
    permission_classes = []

    def get(self, request: Request) -> Response:
        # TODO: This needs to be a bit more flexible if the user wants to change something
        return Response(
            {
                "authentication": get_bridger_auth(request),
                "profile": reverse("bridger:profile", request=request),
                "menu": reverse("bridger:menu", request=request),
                "notification": get_notification_config(request=request),
            }
        )



from rest_framework import serializers

from .mixins import BridgerSerializerFieldMixin
from .types import BridgerType, ReturnContentType


class ImageField(BridgerSerializerFieldMixin, serializers.ImageField):
    field_type = BridgerType.IMAGE.value


class FileField(BridgerSerializerFieldMixin, serializers.FileField):
    field_type = BridgerType.FILE.value



from dataclasses import asdict, dataclass, field
from enum import Enum
from typing import List, Optional


class ButtonLevel(Enum):
    SUCCESS = "success"
    ERROR = "error"
    WARNING = "warning"
    LINK = "link"
    DEFAULT = "default"


class ButtonType(Enum):

    # Buttons
    REFRESH = "refresh"
    NEW = "new"
    DELETE = "delete"

    # Buttons and Create Buttons
    SAVE = "save"
    SAVE_AND_CLOSE = "saveandclose"
    SAVE_AND_NEW = "saveandnew"

    # Create Buttons
    RESET = "reset"

    # Custom Buttons
    DROPDOWN = "dropdown"
    HYPERLINK = "hyperlink"
    WIDGET = "widget"
    ACTION = "action"

    @classmethod
    def buttons(cls):
        return [
            cls.REFRESH.value,
            cls.NEW.value,
            cls.DELETE.value,
            cls.SAVE.value,
            cls.SAVE_AND_CLOSE.value,
            cls.SAVE_AND_NEW.value,
        ]

    @classmethod
    def create_buttons(cls):
        return [
            cls.SAVE.value,
            cls.SAVE_AND_CLOSE.value,
            cls.SAVE_AND_NEW.value,
            cls.RESET.value,
        ]

    @classmethod
    def custom_buttons(cls):
        return [
            cls.DROPDOWN.value,
            cls.HYPERLINK.value,
            cls.WIDGET.value,
            cls.ACTION.value,
        ]


@dataclass
class CustomButton:
    error_messages = {
        "xor_key_endpoint": "Either key or endpoint has to be defined.",
        "or_label_icon": "Either label or icon has to be defined.",
        "exists_button_type": "The field button_type has to exist.",
    }

    level: ButtonLevel = ButtonLevel.DEFAULT

    key: Optional[str] = None
    endpoint: Optional[str] = None
    label: Optional[str] = None
    icon: Optional[str] = None
    title: Optional[str] = None

    def __post_init__(self):
        assert bool(self.key) != bool(self.endpoint), self.error_messages[
            "xor_key_endpoint"
        ]
        assert self.label or self.icon, self.error_messages["or_label_icon"]
        assert hasattr(self, "button_type"), self.error_messages["exists_button_type"]

    # def __dict__(self):
    #     return super().__dict__

    def __iter__(self):
        for field_key in ["key", "endpoint", "label", "icon", "title"]:
            value = getattr(self, field_key, None)
            if value:
                yield field_key, value

        yield "type", self.button_type.value


@dataclass
class HyperlinkButton(CustomButton):
    button_type: ButtonType = ButtonType.HYPERLINK


@dataclass
class WidgetButton(CustomButton):
    button_type = ButtonType.WIDGET


@dataclass
class DropDownButton(CustomButton):
    button_type = ButtonType.DROPDOWN
    buttons: List[CustomButton] = field(default_factory=list)

    def __post_init__(self):
        assert not self.key and not self.endpoint
        assert self.label or self.icon, self.error_messages["or_label_icon"]
        assert hasattr(self, "button_type"), self.error_messages["exists_button_type"]

    def __iter__(self):
        yield from super().__iter__()
        yield "buttons", [dict(button) for button in self.buttons]


btn = DropDownButton(label="A", icon="B", buttons=[HyperlinkButton(key="A", icon="B")])

print(dict(btn))



from typing import Dict, Tuple

from django.conf import settings
from rest_framework.request import Request
from rest_framework.reverse import reverse

from bridger.config import config_registry


# @config_registry.register
def get_notification_config(request: Request):
    scheme = "wss" if request.scheme == "https" else "ws"
    host = request.get_host()
    default_websocket_url = f"{scheme}://{host}/ws/notification/"
    websocket_url = getattr(settings, "BRIDGER_WEBSOCKET_URL", default_websocket_url)

    return {
        "websocket": websocket_url,
        "http": reverse("bridger:notification-list", request=request),
    }



from bridger import serializers

from .models import Notification


class NotificationModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = Notification
        fields = read_only_fields = (
            "id",
            "recipient",
            "title",
            "message",
            "buttons",
            "timestamp_created",
            "timestamp_received",
            "timestamp_read",
            "timestamp_mailed",
        )



from collections import defaultdict
from enum import Enum
from typing import Any, List, NamedTuple, Optional, Tuple, Union

from bridger.display import InstanceDisplay
from bridger.enums import Button


class ButtonLevel(Enum):
    SUCCESS = "success"
    ERROR = "error"
    WARNING = "warning"
    LINK = "link"
    DEFAULT = "default"


class CustomButton(NamedTuple):
    key: Optional[str] = None
    endpoint: Optional[str] = None

    label: Optional[str] = None
    icon: Optional[str] = None
    title: Optional[str] = None

    level: ButtonLevel = ButtonLevel.DEFAULT

    def to_dict(self):
        assert self.label or self.icon, "At least a label or an icon is needed."
        assert self.key or self.endpoint, "At least a key or an endpoint is needed."
        assert self.button_type is not None, "A button type has to be provided."

        rv = defaultdict(dict, {"type": self.button_type, "level": self.level.value})

        fields = ["key", "endpoint", "label", "icon", "title"]

        for field in fields:
            _field = getattr(self, field, None)
            if _field:
                rv[field] = _field

        return rv


class HyperlinkButton(CustomButton):
    button_type = Button.HYPERLINK.value


class WidgetButton(CustomButton):
    button_type = Button.WIDGET.value


class DropdownButton(NamedTuple):
    buttons: List
    label: Optional[str] = None
    icon: Optional[str] = None
    title: Optional[str] = None
    button_type = Button.DROPDOWN.value

    level: ButtonLevel = ButtonLevel.DEFAULT

    def to_dict(self):
        rv = defaultdict(list, {"type": self.button_type, "level": self.level.value})

        for button in self.buttons:
            rv["buttons"].append(button.to_dict())

        if self.label:
            rv["label"] = self.label

        if self.icon:
            rv["icon"] = self.icon

        if self.title:
            rv["title"] = self.title

        return rv


class AdditionalButtonConfig(NamedTuple):
    icon: Optional[str] = None
    label: Optional[str] = None
    title: Optional[str] = None
    level: ButtonLevel = ButtonLevel.DEFAULT

    def to_dict(self):
        rv = dict({"level": self.level.value})

        for field in ["icon", "title", "label"]:
            _field = getattr(self, field, None)
            if _field:
                rv[field] = _field

        return rv


class ActionButton(NamedTuple):
    method: str
    action_label: str

    key: Optional[str] = None
    endpoint: Optional[str] = None

    label: Optional[str] = None
    icon: Optional[str] = None
    title: Optional[str] = None

    description_fields: List[str] = []
    instance_display: Optional[InstanceDisplay] = None
    confirm_config: Optional[AdditionalButtonConfig] = None
    cancel_config: Optional[AdditionalButtonConfig] = None

    identifiers: List[str] = None

    button_type = Button.ACTION.value

    def to_dict(self):
        assert bool(self.key) != bool(
            self.endpoint
        ), "Either key or endpoint can be defined."
        rv = defaultdict(list, {"type": self.button_type})

        rv["method"] = self.method

        if self.endpoint:
            rv["endpoint"] = self.endpoint

        if self.key:
            rv["key"] = self.key

        rv["action_label"] = self.action_label

        rv["description_fields"] = self.description_fields
        if self.instance_display:
            rv["form_display"] = self.instance_display.to_dict()
        else:
            rv["form_display"] = []

        if self.label:
            rv["label"] = self.label

        if self.icon:
            rv["icon"] = self.icon

        if self.title:
            rv["title"] = self.title

        if self.confirm_config:
            rv["confirm_config"] = self.confirm_config.to_dict()

        if self.cancel_config:
            rv["cancel_config"] = self.cancel_config.to_dict()

        if self.identifiers is not None:
            rv["identifiers"] = self.identifiers

        return rv



from collections import defaultdict
from typing import Any, Callable, Dict, List, NamedTuple, Optional

from django.contrib import admin
from rest_framework.request import Request
from rest_framework.reverse import reverse


class ItemPermission(NamedTuple):
    permissions: List[str] = []
    method: Optional[Callable] = None
    # is_employee: bool = False

    def has_permission(self, request: Request):
        if request.user.is_superuser:
            return True

        for permission in self.permissions:
            if not request.user.has_perm(permission):
                return False

        # TODO: This has to be transformed to a employee method permission
        # example: def employee_perm(request) -> return request.user.profile.is_employee
        # if self.is_employee and not request.user.profile.is_employee:
        #     return False

        if self.method:
            return self.method(request=request)

        return True


class PresetMenuItem(NamedTuple):
    label: str
    preset: str
    permission: Optional[ItemPermission] = None

    def to_dict(self, request: Request) -> Dict:
        if self.permission is None or self.permission.has_permission(request):
            return {"label": self.label, "preset": self.preset}
        return None


class MenuItem(NamedTuple):
    label: str
    endpoint: str
    permission: Optional[ItemPermission] = None
    add: Any = None

    def to_dict(self, request: Request) -> Dict:
        if self.permission is None or self.permission.has_permission(request):
            item = {
                "label": self.label,
                "endpoint": reverse(viewname=self.endpoint, request=request),
            }

            if self.add:
                item["add"] = self.add.to_dict(request=request)

            return item

        return None


class Menu(NamedTuple):
    label: str
    items: List

    def to_dict(self, request: Request) -> Dict:
        menu_dict = {"label": self.label, "items": list()}
        for item in self.items:
            _item = item.to_dict(request=request)
            if _item:
                menu_dict["items"].append(_item)

        if len(menu_dict["items"]) == 0:
            return None

        return menu_dict


class MenuRegistry:
    def __init__(self):
        self._registry = list()

    def register(self, menu: Menu):
        self._registry.append(menu)

    def to_dict(self, request: Request) -> List:
        menu_list = list()
        for menu in self._registry:
            _menu = menu.to_dict(request=request)
            if _menu is not None:
                menu_list.append(_menu)
        return menu_list


default_registry = MenuRegistry()

# default_registry.register(PresetMenuItem(label="News", preset="StainlyNewsWidget"))

# default_registry.register(
#     Menu(
#         label="Our Menu",
#         items=[
#             MenuItem(
#                 label="M1",
#                 endpoint="e1",
#                 add=MenuItem(label="Add Something", endpoint="ea1"),
#             ),
#             Menu(label="Nested Menu", items=[MenuItem(label="N-M1", endpoint="e1"),],),
#         ],
#     )
# )

# import pprint

# pprint.pprint(default_registry.to_dict(request=None))



from enum import Enum

import celery
from django.contrib.auth import get_user_model
from django.contrib.postgres.fields import JSONField
from django.db import models, transaction
from django.db.models.signals import post_save
from django.dispatch import receiver

from bridger.buttons import CustomButton, WidgetButton


class NotificationSendType(Enum):
    SYSTEM = "system"
    MAIL = "mail"
    SYSTEM_AND_MAIL = "system_and_mail"


class Notification(models.Model):
    """
        A model that represents a notification that is send as a system notification or mail or both
    """

    recipient = models.ForeignKey(
        to=get_user_model(), related_name="notifications", on_delete=models.CASCADE
    )

    title = models.CharField(max_length=512)
    message = models.TextField(null=True, blank=True)
    buttons = JSONField(default=list, null=True, blank=True)

    timestamp_created = models.DateTimeField(auto_now_add=True, verbose_name="Created")
    timestamp_received = models.DateTimeField(
        null=True, blank=True, verbose_name="Received"
    )
    timestamp_read = models.DateTimeField(null=True, blank=True, verbose_name="Read")
    timestamp_mailed = models.DateTimeField(
        null=True, blank=True, verbose_name="Mailed"
    )

    send_type_choices = (
        (NotificationSendType.SYSTEM.value, "System"),
        (NotificationSendType.MAIL.value, "Mail"),
        (NotificationSendType.SYSTEM_AND_MAIL.value, "System and Mail"),
    )

    send_type = models.CharField(
        max_length=32,
        choices=send_type_choices,
        default=NotificationSendType.SYSTEM.value,
    )

    def __str__(self):
        return f"{self.recipient} {self.title}"

    def to_payload(self):
        return {
            "id": self.id,
            "title": self.title,
            "message": self.message,
            "buttons": self.buttons,
        }


@receiver(post_save, sender=Notification)
def post_create_notification(sender, instance, created, **kwargs):
    send_mail = "bridger.notifications.send_mail"
    send_system = "bridger.notifications.send_system"
    if created:
        dispatch = {
            NotificationSendType.MAIL.value: [send_mail],
            NotificationSendType.SYSTEM.value: [send_system],
            NotificationSendType.SYSTEM_AND_MAIL.value: [send_system, send_mail],
        }
        for action in dispatch[instance.send_type]:
            transaction.on_commit(
                lambda: celery.execute.send_task(action, args=[instance.id])
            )



from typing import Callable, Iterable


class BridgerConfig:
    def __init__(self):
        self._config_list = []

    def register(self, func: Callable):
        self._config_list.append(func)

    def serialize(self, request) -> Iterable:
        for func in self._config_list:
            yield func(request)


config_registry = BridgerConfig()


@config_registry.register
def do_something():
    return "notification", {"hello": "world"}


# print(config_registry.serialize())

# for k in config_registry:
#     print(k)
